
// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.


#import "AVLSet.h"


@implementation AVLSet

//compare functions for AVLtrees must take 3 args, 2 objects and 1 optional parameter.
static int
compare_object_ids (const void *A, const void *B, void *PARAM)
{
  int a = (unsigned) A;
  int b = (unsigned) B;

  if (a  < b )
    return -1;

  return a > b;
}

static int
compare_integers (const void * val1, const void *val2, void *PARAM)
{
  if ((PTRINT) val1 < (PTRINT) val2)
    return -1;
  return ((PTRINT) val1 > (PTRINT) val2);
}

static int
compare_unsigned (const void * val1, const void * val2, void * PARAM)
{
  if ((PTRUINT) val1 < (PTRUINT) val2)
    return -1;
  return ((PTRUINT) val1 > (PTRUINT) val2);
}

static int
compare_cstrings (const void* val1, const void * val2, void * PARAM)
{
  return strcmp ((const char *)val1, (const char *)val2);
}

PHASE(Creating)

+createBegin: aZone  
{
  AVLSet * obj; 
  obj = [super createBegin: aZone];
  return obj;
}


- setReplaceOnly: (BOOL) condition
{
  replaceOnly = condition;
  return self;
}

//It is mandatory to choose one of these compare functions at create time.

- setCompareFunction: (compare_avl_t) compareFunction
{
  compareFunc = compareFunction;
  aTree = avl_create ( compareFunc, NULL);
  return self;
}

- setCompareCStrings
{
  compareFunc = compare_cstrings;
  aTree = avl_create ( compareFunc, NULL);
  return self;
}

- setCompareIDs
{
  compareFunc = compare_object_ids;
  aTree = avl_create ( compareFunc, NULL);
  return self;
}

- setCompareIntegers
{
  compareFunc = compare_integers;
  aTree = avl_create ( compareFunc, NULL);
  return self;
}

- setCompareUnsignedIntegers
{
  compareFunc = compare_unsigned;
  aTree = avl_create ( compareFunc, NULL);
  return self;
}

//by default,use IDs to compare. That makes it a Set.
-createEnd
{
  replaceOnly = NO;
  if ( aTree == NULL )
    raiseEvent (InvalidArgument, "You Forgot to set a compare function");
  return [super createEnd];
}

//convenience method
+createBegin: (int (*)(const void*,const void*,void*)) comparFunc  Param: (void *) param Zone: aZone
{
  AVLSet * obj; 
  obj = [super createBegin: aZone];
  obj->aTree = avl_create (comparFunc, param);
  return obj;
}

PHASE(Using)

//Methods in Set certainly have to be implemented here, so here they are:

- copy: aZone
{
  AVLSet *newSet;
  newSet = [aZone copyIVars: self];
  newSet->aTree = avl_copy (aTree,NULL);
  return newSet;
}

- (int) count
{
  return avl_count(aTree);
}

- (unsigned) getCount
{
  return (unsigned)  avl_count(aTree);
}



//add an object if it is not currently in the set
//if anObject is in the set, then return that value.
//does nothing and returns NO otherwise
-(BOOL) add: anObject
{
  id value = avl_insert (aTree, anObject ); //uses integer keys;
  if (value == NULL)
    return YES;
  else return NO;
}

// if anObject matching key exists, replace it, return old one.
// otherwise do nothing and return nil
- (id *) replace: anObject
{
  void * value;
  if (( value = avl_find(aTree, anObject) ))
    {
      avl_replace (aTree, anObject);
    }
  return value; //have to worry about NULL being different from nil?
}

- (BOOL) getReplaceOnly
{
  return replaceOnly;
}


- (BOOL) contains: aKey
{
  if (avl_find(aTree, aKey))
    return YES;
  else
    return NO;
}

- (BOOL)containsKey: aKey
{
  return [self contains: aKey];
}


- at: aKey
{
  void * value=avl_find(aTree, aKey);
   return value;
}


- (BOOL)at: aKey memberSlot: (id **)memptr
{
  raiseEvent (NotImplemented, nil);
  return NO;
}

- remove: aKey
{
  return avl_delete(aTree, aKey);
}

- removeKey: aKey
{
  return avl_delete(aTree, aKey);
}

- (void)forEach: (SEL)aSelector
{
  id anObject;
   avl_traverser trav = {0};
   while (( anObject = avl_traverse (aTree, &trav)) != NULL) 
     [anObject perform: aSelector];
}

- (void)forEach: (SEL)aSelector : arg1
{
  id anObject;
   avl_traverser trav = {0};
   while (( anObject = avl_traverse (aTree, &trav)) != NULL )
     [anObject perform: aSelector with: arg1];
}

- (void)forEach: (SEL)aSelector : arg1 : arg2
{ 
  id anObject;
   avl_traverser trav = {0};
   while (( anObject = avl_traverse (aTree, &trav)) != NULL )
     [anObject perform: aSelector with: arg1 with: arg2];
}

- (void)forEach: (SEL)aSelector : arg1 : arg2 : arg3
{
  id anObject;
   avl_traverser trav = {0};
   while (( anObject = avl_traverse (aTree, &trav)) != NULL )
     [anObject perform: aSelector with: arg1 with: arg2 with: arg3];
}




- (void)forEachKey: (SEL)aSelector
{
  [self forEach: (SEL)aSelector];
}
- (void)forEachKey: (SEL)aSelector : arg1
{
  [self forEach: (SEL)aSelector: arg1];
}
- (void)forEachKey: (SEL)aSelector : arg1 : arg2
{
  [self forEach: (SEL)aSelector: arg1 : arg2];
}
- (void)forEachKey: (SEL)aSelector : arg1 : arg2 : arg3
{
  [self forEach: (SEL)aSelector: arg1 : arg2];
}



- (void)describeForEach: outputCharStream
{
  id index, member;
  index = [(id) self begin: [self getZone]];
  for (member = [index next];
       [index getLoc] == Member;
       member = [index next])
    [member describe: outputCharStream];
  [index drop];
}

- (void)describeForEachID: outputCharStream
{
  id index, member;
  index = [(id) self begin: [self getZone]];
  for (member = [index next];
       [index getLoc] == Member;
       member = [index next])
    [member describeID: outputCharStream];
  [index drop];
}




-begin: aZone
{
  AVLSetIndex_c * newIndex;

  newIndex = [aZone allocIVars: [AVLSetIndex_c self]];
  newIndex->tree = aTree;
  //  newIndex->collection = self;
  [newIndex init];
  return newIndex;
}



- createIndex: (id <Zone>)aZone fromMember: anObject 
{
  raiseEvent(WarningMessage,"Not Implemented");
  return self;
}


- (void) drop
{
  avl_destroy(aTree,NULL); //does not drop objects tree refers to
  [super drop];
}

- (void) removeAll
{
  avl_destroy(aTree,NULL);
}


- (void) deleteAll
{ 
  void dropObjects(void * data, void * param)
    {
      [(id) data drop];
    }
  avl_destroy(aTree,dropObjects);
  //avl_free(aTree);  //drops objects in the tree and tree too
}



// Super power insert. Inserts if object does not exist, replaces if object does.
// returns object being replaced if there is one.
- forcefulReplace: anObject
  {
    id value = avl_replace (aTree, anObject );//uses integer keys;
    if (value != NULL) 
      return value;
    else return nil;
  }


-(void)  walk: (void (*)(void*,void*)) aFunction Param: (void *) PARAM
{
  avl_walk(aTree , aFunction, PARAM);
}


- traverse: (avl_traverser *) travPtr
{
  id anObject;

  anObject = avl_traverse (aTree, travPtr) ;
  
 return anObject;
}

- (avl_tree *) getTree
{
  return aTree;
}


- atOffset: (unsigned) x
{
  id index = [self begin: [self getZone]];
  id item;
  if ( x > (unsigned)avl_count(aTree)-1)
    [InvalidArgument raiseEvent: "Bad Offset in access of AVL tree \n" ];

  item = [index next];
  while ( (unsigned)[index getOffset] < x )
    item = [index next];
  [index drop];
  return item;
}

- atOffset: (unsigned) x put: anObject
{

  [InvalidArgument  raiseEvent: "atOffset:put: makes no sense in a tree \n"];
  return self;
}

- getFirst
{
  return [self atOffset: 0];
}

- getLast
{
  return [self atOffset: avl_count(aTree)-1];
}

- removeFirst
{
  id item = [self atOffset: 0];
  avl_delete(aTree, item );
  return item;
}

- removeLast
{
  id item = [self atOffset: avl_count(aTree)-1];
  avl_delete(aTree, item );
  return item;

}


- (void)setInitialValue: initialValue 
{
  [InvalidArgument raiseEvent: "setInitialValue not implemented for trees\n"];
}



- beginPermuted: aZone
{ 
  raiseEvent (NotImplemented, nil);
  return self;
}

- (BOOL) allSameClass
{ 
  raiseEvent (NotImplemented, nil);
  return NO;
}

- (void) setIndexFromMemberLoc: (int)byteOffset 
{ 
  raiseEvent (NotImplemented, nil);
}


@end



@implementation AVLSetIndex_c
- next
{
  currentObject =  avl_traverse (tree, &trav);
  offset ++;
  return  currentObject ; //returns a null at the end of the tree
}

//sorry, I don't see a way to make an avltree walk backwards!
- prev
{
  int i;
  int currentOffset = offset;
  void * anObject = NULL;

  [self init];
  for (i=0; i < currentOffset; i++)
    {
      anObject = [self next];
    }
  printf("previousOffset was %d, new offset is %d \n",currentOffset, offset);
  return anObject;
}

- get
{
  return currentObject;
}

- replace: anObject
{
  void * value;
  if (( value = avl_find(tree, anObject) ))
    {
      avl_replace (tree, anObject);
    }
  return value;  //removed object
}

- remove
{
  id removed;
  int currentOffset = offset;
 
  removed = avl_delete ( tree,  currentObject );  
  [self init];
  while ( offset < currentOffset )
    {
      [self next];  //step up to object before removal
    }
  return removed;
}


- init
{
  avl_traverser newtrav = {0};
  trav = newtrav ;
  offset = -1;
  currentObject = nil;
  return self;
}

- (id <Symbol>)getLoc
{
  if (offset == -1) return Start;
  else if (currentObject != NULL) 
    {
      fprintf (stderr,"Member! offset= %d",offset);
      return Member;
    }
  else if (offset >= avl_count(tree))
    {
      return End;
    }
  else
    raiseEvent(WarningMessage,"Invalid AVLSet position");
  return End;
}

- (void)setLoc: (id <Symbol>)locSymbol
{
  if ( locSymbol == Start)
    [self init];
  if ( locSymbol == End)  //this is silly, nobody should ever do it. 
    while (avl_traverse(tree, &trav));
}

- (int)getOffset
{
  return offset ;
}

- setOffset: (unsigned)targetOffset
{
  [self init];
  do
    {
      [self next];
    } while ( (unsigned)offset < targetOffset );
  return  currentObject;
}

//I don't know what this does!
//  - (void) mapAllocations: (mapalloc_t)mapalloc
//  {
//    mapObject (mapalloc, listIndex);
//  }

//Methods from "Index_any" that do not have meaning in a Set

- findNext: anObject
{
  raiseEvent (NotImplemented, nil);
  return self;
}

- findPrev: anObject
{
  raiseEvent (NotImplemented, nil);
  return self;
}

-getCollection
{
  return (id) tree;
}

- put: anObject
{
  id oldMem;
  
  oldMem = avl_replace (tree, anObject);  
  if (oldMem == nil) raiseEvent (InvalidIndexLoc, nil);
  return oldMem;
}

-_I_Using
{
  raiseEvent (NotImplemented,nil);
  return self;
}

@end
