// BatchSwarm.m


// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.

#import "MyArguments.h"
#import "BatchSwarm.h"
#import "ModelSwarm.h"
#import "MultiGrid2d.h" //needs "passThrough" to get drawing
#import <simtools.h> // ObjectLoader

#import <float.h>  //for "DBL_MIN"
#import "analysis/Averager.h"


#include <misc.h> // printf

@implementation BatchSwarm

// createBegin: here we set up the default observation parameters.

+ createBegin:  aZone 
{
  BatchSwarm *obj;

  // Superclass createBegin to allocate ourselves.
  obj = [super createBegin: aZone];

  // Fill in the relevant parameters.
  obj->displayFrequency = 1;
  obj->experimentDuration = getInt(arguments,"duration");

  return obj;
}

- buildObjects
{

  char hdfEZGraphName[100];

  id hdf5container;
  [super buildObjects];
  modelSwarm = [ModelSwarm create: self];
  [modelSwarm buildObjects];

  sprintf (hdfEZGraphName, "latane%d.hdf",[(id)arguments getRunArg]);
  
  hdf5container = [HDF5 createBegin: [self getZone]];
  [hdf5container setWriteFlag: YES];
  [hdf5container  setName: hdfEZGraphName];
  hdf5container = [hdf5container createEnd];

  //  if(displayFrequency)
  //  {
      opinionGraph = [EZGraph createBegin: self];
      [opinionGraph setGraphics: 0] ;
      [opinionGraph setFileOutput: 1] ;
      [opinionGraph setHDF5Container: hdf5container];
      //in hdf5 output, "File" is really the prefix used inside the file
      //name of file is value of hdf5EZGraphName
      [opinionGraph setFileName: "latane" ];
      opinionGraph = [opinionGraph createEnd] ;
      // }

  printf("Create Sequences");
  [opinionGraph createSequence: "OpinionYES"
			       withFeedFrom:  [modelSwarm getInfluenceGrid]
			       andSelector: M(getPctTypeOne)];
  
  sequence = [opinionGraph createAverageSequence: "Opinion Changed"
			       withFeedFrom: [modelSwarm getCitizenList]
			       andSelector: M(getChanged)];

  
  [opinionGraph createAverageSequence: "Bolstered Minority"
			       withFeedFrom: [modelSwarm getCitizenList]
			       andSelector: M(getMinority)];

  //now setup raster to write
  {
    int i;
    colormap = [BatchColormap create: self];
    for (i = 0; i < 101; i++)
      [colormap setColor: i ToRed: i/ 105.0  Green: 0 Blue: 4.0/64.0];

  // Colour 64 is set to green, to display heatbugs

    [colormap setColor: 102 ToName: "green"];

  // Colour 65 is set to white, used in this case below on probed heatbug.

    [colormap setColor: 103 ToName: "white"];

    for(i=104; i < 169; i++)
      [colormap setColor: i  ToRed: 0  Green: (double)(i-104) / 65.0 Blue: 0.4];
    
    {
      //      char rasterName[20];
      //   sprintf(rasterName,"SI");
      worldRaster = [BatchRaster createBegin: self];
      worldRaster = [worldRaster createEnd];
      [worldRaster setColormap: colormap];
      [worldRaster setWidth: getInt(arguments,"worldXSize")
		   Height: getInt(arguments,"worldYSize")];
      [worldRaster pack];     // draw the window.
      

      //if we had a regular raster we could do this:
      //   opinionDisplay = [Object2dDisplay createBegin: self]; 
      //       opinionDisplay = [Object2dDisplay create: self
      // 					setDisplayWidget: worldRaster
      // 					setDiscrete2dToDisplay: [modelSwarm getPositionGrid]
      // 					setDisplayMessage: M(drawSelfOn:)];
    }
  }
return self;
}  

// Create the actions necessary for the simulation. This is where
// the schedule is built (but not run!)

- buildActions 
{
  [super buildActions];
  
  // First, let our model swarm build its own schedule.

  [modelSwarm buildActions];
  
  if(displayFrequency)
    {
      displayActions = [ActionGroup create: self];
      
      [displayActions createActionTo: opinionGraph message: M(step)];
      
      displaySchedule = [Schedule createBegin: self];
      [displaySchedule setRepeatInterval: displayFrequency];
      displaySchedule = [displaySchedule createEnd];
      
      [displaySchedule at: 0 createAction: displayActions];
    }
  
  [displaySchedule at: 0 createActionTo: self message: M(checkToStop)];

  return self;
}  

// activateIn: - get the Swarm ready to run.
- activateIn:  swarmContext 
{
  // First, activate ourselves (just pass along the context).
  [super activateIn: swarmContext];
  
  [modelSwarm activateIn: self];
  
  if (displayFrequency)
    [displaySchedule activateIn: self];
 
  return [self getActivity];
}


- go 
{
  //  printf ("You typed `-b', so we're running without graphics.\n");

  if (displayFrequency)
      //      printf ("It is display data every %d timesteps to: unhappiness.output.\n",

  [[self getActivity] run];

  return [[self getActivity] getStatus];
}


- checkToStop
{
  double myAvg =99 ;
  long time = getCurrentTime();
   myAvg = [sequence getCurrentValue];
   
   printf ("average is %f \n", myAvg);
  //DBL_MIN: a secret Marcus told me about from ISO C: smallest value of double
   if ( ((myAvg < DBL_MIN) && (time > 5)) || ((time  > experimentDuration)) )
     {
       // printf ("avg equalz 0 \n");
       [self unattendedStop];
     }

   return self;
}

- unattendedStop 
{
  char filename[25]; 
  printf ("unattended stop: monitored value says it is time to quit: There were no changes in opinion\n");

  MultiGrid2d * aGrid =[modelSwarm getPositionGrid];

  [aGrid passThroughToAllCells: M(drawSelfOn:) with: worldRaster];

  sprintf(filename,"opinion-%d-%ld.png",getInt(arguments,"run"),getCurrentTime());
  [worldRaster  writeSelfToFile: filename];  
 
  printf ("%04d RUN %07ld timesteps\n", [(id)arguments getRunArg],getCurrentTime ());
  
  [modelSwarm finishLog]; 

//   if(displayFrequency){
//     [opinionGraph drop] ;              // Close the output file.
//     opinionGraph = nil;
//   }
  
  [getTopLevelActivity() terminate];  // Terminate the simulation.    
  return self;
}



/*"It is necessary to drop the data writing objects in order to make
sure they finish their work.
"*/
-(void) drop
{
  if (opinionGraph)  [opinionGraph drop];
 //[super drop];
}


@end










