
// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this software and offers no 
// warranty of any kind.

// Class MultiGrid2d.h, Paul Johnson, descended from 
// Class "MoGrid2d" by Sven Thommesen
// which had been modified by Pietro Terna

// Introduction This replacement for Grid2d allows for Multiple
// Occupancy!  Each occupied position has a "cell" object, and that
// cell holds objects a collection.  MultiGrid2d is not a sublclass of
// Grid2d, it is a free standing combination space/container class.
// It is free standing and separate from Grid2d partly to assure that
// no IVARS or methods are unintentinally inherited in this
// multi-agent-per-cell environment.  

// To assure that MultiGrid2d can act as a data provider by Swarm, it
// adopts the GridData protocol of Swarm, which means that it is fully
// qualified to act as a discrete2dToDisplay for Raster displays.

// By default, the cell uses an ordinary Swarm list as the collection,
// but the user can provide for a new kind of container.  Why this
// feature?  Some collections are faster for extracting particular
// objects by name, while others are faster for finding an object at a
// particular offset, while others are faster for finding the oldest,
// biggest, fattest, and so forth. If your objects are fixed at
// particular positions, a cell that is based on a Swarm Array would
// work great.  So I generalize to allow any kind of collection the
// user desires. I also allow user-defined methods to be accessed for
// any particular position in the grid.

// How is this done?  Think of each "thing" in the grid as a cell,
// which can accept and give back items, and it can also carry out
// other actions you want.  The "default" cell class is MultiGridCell,
// which uses a swarm list class as a container.  The user can create
// a subclass of MultiGridCell, or can create a class from scratch, as
// long as it adopts the protocol AppCell, which is defined in
// MultiGridCell.h.  This protocol requires that the cell container be
// able to add, remove, and find objects, and also to pass through
// messages to them.  My AppCell (documented in ) uses Swarm
// collection style names for methods.  All of this seems pretty
// obvious to me, but only because i've been looking at this for 3
// years.

// When the grid is created, the user must tell the grid what "cell
// Class" to use.  If none is specified before createEnd is called,
// then the MultiGridCell is assumed.  (The setCellClass: method 
// sets the cell class for the grid.)

// The Cell object has a list instance variable. That's the
// collection. It may also have other instance variables and it can be
// used for any data-gathering purpose, including display purposes
// (create drawSelfOn method and so forth).  Objects/agents are added
// to or removed from such lists, rather than from the underlying
// grid.  Cells are created as needed, dynamically, when agents are
// added at a point.  (The grid is created initially with no Cell
// objects.)  This version uses a recycler to grab MultiCells that are
// no longer in use.  They are initialized and reused.  This behavior
// saves on memory and it may even make a model faster if cells need
// not be constantly recreated.  It depends on whether your grid is
// sparsely populated, I think.  You can turn off the automatic
// removal of empty cells with a CPP flag,
// NO_AGGRESSIVE_RECYCLING. This leaves the cells, so they eagerly
// await a chance to insert an object for you.

// I have many methods in here with names that start with
// "passThrough".  These methods take a selector argument (and
// possibly another argument).  The idea here is that if one wants
// each cell object to carry out a certain method, then the pass
// through object can pass through that selector's method to all cells
// (see passThroughToAllCells: ). Or, in the case of
// passThroughToAllObjects, it causes each object in the grid to be
// sent a message.  Or, in the cases of passThroughToCell:atX:Y:, or
// passThroughToObjects:atX:Y:, a message is passed to a particular
// cell or to all objects in that cell.  If you have ever looked at
// the Objective-C root class Object, you will have seen the "perform:
// (SEL) aSelector" method, and I use that same idea in these pass
// through methods.

// Because of these "passThrough" methods, users who create a custom
// cell class have to be a little careful to implement or at least
// create stubs for methods like "performForEach". Or subclass from
// MultiGridCell.  Since MultiGridCell implements these, I believe
// that any subclass that uses a Swarm compliant collection will be
// acceptable and these methods need not be overridden.

// There is a special passThrough method, though, which I would point
// out:

/*
- passThroughToCellCreation: (SEL) aSelector{
*/

// This sets the values of features that might be needed when cells
// are created inside the grid.  If your Cell class requires some
// customization, say a Cell object needs to have a variable set
// or something like that, then this passThrough makes sure the Grid
// knows what it is supposed to do.  It works so that
// every time the Grid creates a cell, it performs this method, i.e.,

/*
  [cell perform: aSelector];
*/

// I have found some cases where there is no other way to get the
// desired functionality of the Cell and still work with a general
// purpose grid like MultiGrid2d.


// Here is an example of a sequence of commands that create a
// MultiGrid2d with a user-customized cell class called "AppSpecificCell".

/*

homeGrid = [ MultiGrid2d create: self setSizeX: worldXSize Y: worldYSize Cells: [AppSpecificCell class]];
[homeGrid passThroughToCellCreation: M(setRasterList:) with: [observerSwarm getRasterList: nhood]];
*/ 


// The grid is created, told to use cells of class AppSpecificCell,
// and every time a cell is created, it is told to setRasterList: with
// a given input variable.  So that whenever it spawns a new cell, it
// passes on a particular message to that cell. 

// Sometimes I have found it handy to have a "global_list" of all
// objects that have ever been inserted.  In MultiGrid2d.[hm], one can
// see CPP flags that turn on the creation of a global object list and
// maintenance of it when items are inserted and removed.  This
// version uses an AVL Tree to keep track of objects so that removal
// of specific objects is fast.  This big list can be used to tell all
// objects in the grid to do something, or to get a count of all
// objects in the grid. Of course, you need my AVLSet class to make
// that work, but it is included in this tarball. The CPP flag for
// that is KEEP_GLOBAL_LIST

// I'm also providing an example of a cell subclass that I use in an
// application.

// The original MoGrid2d was authored by Sven Thommesen (c) Sven
// Thommesen 1997 <sthomme@humsci.auburn.edu>.  Sven's grid held Swarm
// lists.  It had some functionality to copy and shuffle lists that I
// don't maintain here.  If you need to shuffle a list, get the
// collection and shuffle it yourself.

// Sven is not responsible for any errors in this new (significantly
// revised and generalized) multiple occupancy grid.

// My earlier effort WrappedListGrid2d was not so generally functional
// as this one.

// (c) Paul Johnson 2001 <pauljohn@ukans.edu>

#import <objectbase/SwarmObject.h>
//#import <space/Grid2d.h>
#import "MultiGridCell.h"
#import <space.h>
#import <defobj.h>
#import <defobj/Create.h>

// To create a list of all inserted objects, set the CPP flag
// KEEP_GLOBAL_LIST

#import "AVLSet.h"


@interface MultiGrid2d: SwarmObject <GridData> {
  //@interface MultiGrid2d: Grid2d {
  unsigned xsize, ysize;
  
  id *lattice;
  long *offsets;

  id <List> recyclerList;
  Class cellClass; 
  SEL  cellSelector1, cellSelector2;
  id cellArgument;

#ifdef KEEP_GLOBAL_LIST 
  AVLSet * globalList;  //all objects
#endif 
}



-makeOffsets;

-(id *) allocLattice;

-setSizeX: (unsigned)x Y: (unsigned)y;

-(void) setCellClass: (Class)aClass;

-createEnd;

// Creation: If you don't specify a class here, it assumes you
// want ordinary Swarm lists as your containers in the cells.
// If you custom-design a cell class which has containers and
// implements the protocol "AppCell", then use your cell instead.

+create: (id)aZone setSizeX: (unsigned)x Y: (unsigned)y;
+create: (id)aZone setSizeX: (unsigned)x Y: (unsigned)y Cells: (Class)aClass;


// These are copied from Swarm's Grid2d class. These are included in
// order to make this class "free standing", not a subclass of Grid2d.

-(unsigned)   getSizeX;		// x size
-(unsigned)   getSizeY;		// y size
-(id *)  getLattice;		// pointer to the lattice itself
-(long *) getOffsets;            //pt long with Grid2d  inheritance

// Methods to add objects to cells on the lattice: Dump in an
// object. Note it does not say "addFirst" or "addLast", just add, and
// the CELL class has an "insertItem" method which the USER designs to
// add according to any principle.  If the collection is a Hash, add
// has a different meaning than a Swarm List, for example.

//These two do the same thing, incidentally.

- add: (id) anObject atX: (unsigned)x Y: (unsigned)y;
- putObject: anObject atX: (unsigned)x Y: (unsigned)y;//pj:vital to override Grid2d method

// Previous (putObject) and getObject are overrides of Grid2d methods
- getObjectAtX:(unsigned)x Y: (unsigned)y;


- (long)getValueAtX: (unsigned)x Y: (unsigned)y;


//getCell is same as getObject, but more obvious in meaning, in my opinion
- getCellAtX: (unsigned)x Y: (unsigned)y;

-(id) getFirstAtX: (unsigned)x Y: (unsigned)y ;

-(id) getLastAtX: (unsigned)x Y: (unsigned)y ;

// Return the number of objects in the list at (x,y):
-(unsigned) numberOfObjectsAtX: (unsigned)x Y: (unsigned)y;


-(id) getLastAtX: (unsigned)x Y: (unsigned)y;

// Remove and return the object that is "first", according to whatever
// logic the cell's collection class uses.  Or nil if no such object:

-(id) removeFirstAtX:    (unsigned)x Y: (unsigned)y; 

// Remove and return the last object, according to whatever logic your
// collection uses for last (or nil if no such object):

-(id) removeLastAtX:     (unsigned)x Y: (unsigned)y;

// Remove and return the given object from the list at (x,y),
// or (nil) if not found:

-(id) deleteObject: (id) anObject atX: (unsigned)x Y: (unsigned)y;

// Does not remove from the cell, but returns a randomly chosen
// element;
-(id) selectRandomMemberX: (unsigned)x Y: (unsigned)y;

// Remove all items from the cell at (x,y)
// (but the objects themselves still persist, wherever they are):

- (void) removeAllAtX: (unsigned)x Y: (unsigned)y ;

// Remove all object pointers from the list at (x,y) 
// and also drop those objects
- (void) deleteAllAtX: (unsigned)x Y: (unsigned)y;

// Plug the given collection into the cell at (x,y):
// WARNING: we can only have one list per cell!
// setList will abort if there's already a list there.
//-setNewList: list atX: (unsigned) x Y: (unsigned) y;


// Return the collection at (x,y)
-getCollectionAtX: (unsigned)x Y: (unsigned)y;
 
// Get AVLSet collection of all objects in grid
#ifdef KEEP_GLOBAL_LIST
-(AVLSet*) getGlobalList;
#endif

// Cell creation code. Used in many places when items are inserted.
- createCellX: (unsigned)x Y: (unsigned)y;


// Grab random item in list at x y, and open a  probe for it.
- makeProbeAtX: (unsigned)x Y: (unsigned)y;


//save selector to use when creating cells
- passThroughToCellCreation: (SEL)aSelector;

//save selector/argument to use when creating cells
- passThroughToCellCreation: (SEL)aSelector with: arg1;

// Tell all existing cells to execute a method
- passThroughToAllCells: (SEL)aSelector;

- passThroughToAllCells: (SEL)aSelector with: arg1;

// Tell cell at x,y to execute a method
- passThroughToCell: (SEL) aSelector atX: (unsigned)x Y: (unsigned) y;

- passThroughToCell: (SEL)aSelector with: arg1 atX: (unsigned)x Y: (unsigned) y;

// Tell all objects in the grid to do something.
- passThroughToObjects: (SEL)aSelector;

- passThroughToObjects: (SEL)aSelector with: anObject;

// Tell all objects at x,y  to do something.
- passThroughToObjects: (SEL)aSelector atX: (unsigned)x Y: (unsigned) y;

- passThroughToObjects: (SEL)aSelector with: arg1 atX: (unsigned)x Y: (unsigned) y;

//- writeValuesToFile;

- _I_Using;

#define discrete2dSiteAt(l, offsets, x, y) ((l) + (offsets)[(y)] + (x))

@end








