// ObserverSwarm.m

// Paul Johnson <pauljohn@ku.edu> 2002
// This code is made available under the GPL (GNU general public license)
// with the additional proviso that any users notify the author of their
// use and share their code with the original author.  A copy of version
// 2 of the GPL is included in this directory, but you can use a newer
// GPL at your choice.  The author assumes no liability for any 
// inconvenience or damage caused by this code.

#import "ObserverSwarm.h"
#import "ModelSwarm.h"
#import <collections.h>
#import <objectbase.h>
#import <analysis.h>
#import <gui.h>
#import "MultiGrid2d.h"
#import "MyArguments.h"
#import "analysis/Averager.h"
#import <float.h>  //for "DBL_MIN"

@implementation ObserverSwarm

+ createBegin: aZone
{
  ObserverSwarm *obj;
  id <ProbeMap> probeMap;
  
  obj = [super createBegin: aZone];
  
  obj->displayFrequency = 1;
  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];
  
  [probeMap addProbe: [probeLibrary getProbeForVariable: "displayFrequency"
				    inClass: [self class]]];
  
  [probeLibrary setProbeMap: probeMap For: [self class]];

  return obj;
}

- createEnd
{
  return [super createEnd];
}

- _opinionRasterDeath_ : caller
{
  [opinionRaster drop];
  opinionRaster = nil;
  return self;
}

- _pressureRasterDeath_ : caller
{
  [pressureRaster drop];
  pressureRaster = nil;
  return self;
}


- _opinionGraphDeath_ : caller
{
  [opinionGraph drop];
  opinionGraph = nil;
  return self;
}


- buildObjects
{
  int i;
  static const char * ezgraphColors[4]= {"blue", "black", "red", "orange"};
  char hdfEZGraphName[100];
  id <HDF5> hdf5container; //EZGraph data output
  [super buildObjects];
 
  if (getInt (arguments, "UNATTENDED"))
    [controlPanel setStateRunning];
  else
    [controlPanel setStateStopped];
 
  CREATE_ARCHIVED_PROBE_DISPLAY (self);
 
  modelSwarm = [ModelSwarm create: self];

  CREATE_ARCHIVED_PROBE_DISPLAY (modelSwarm);

  [modelSwarm buildObjects];

  colormap = [Colormap create: self];

  // Colours [0,100] are assigned to the range Red [0, 1)

  // [colormap setColor: 0 ToName: "white"];

  for (i = 0; i < 101; i++)
    [colormap setColor: i ToRed: i/ 105.0  Green: 0 Blue: 4.0/64.0];

  // Colour 64 is set to green, to display heatbugs

  [colormap setColor: 102 ToName: "green"];

  // Colour 65 is set to white, used in this case below on probed heatbug.

  [colormap setColor: 103 ToName: "white"];

  for(i=104; i < 169; i++)
    [colormap setColor: i  ToRed: 0  Green: (double)(i-104) / 65.0 Blue: 0.4];

  // Next, create a 2d window for display, set its size, zoom factor, title.
  
  opinionRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME (opinionRaster);
  opinionRaster = [opinionRaster createEnd];
  [opinionRaster enableDestroyNotification: self
		 notificationMethod: @selector (_opinionRasterDeath_:)];
  [opinionRaster setColormap: colormap];
  [opinionRaster setZoomFactor: 4];
  [opinionRaster setWidth: [[modelSwarm getPositionGrid] getSizeX]
		 Height: [[modelSwarm getPositionGrid] getSizeY]];
  [opinionRaster setWindowTitle: "Opinion (YES=green)"];
  [opinionRaster pack];				  // draw the window.


  pressureRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME (pressureRaster);
  pressureRaster = [pressureRaster createEnd];
  [pressureRaster enableDestroyNotification: self
		  notificationMethod: @selector (_pressureRasterDeath_:)];
  [pressureRaster setColormap: colormap];
  [pressureRaster setZoomFactor: 4];
  [pressureRaster setWidth: [[modelSwarm getPositionGrid] getSizeX]
		  Height: [[modelSwarm getPositionGrid] getSizeY]];
  [pressureRaster setWindowTitle: "Net Visible YES"];
  [pressureRaster pack];		


  // Now create a Value2dDisplay: this is a special object that will
  // display arbitrary 2d value arrays on a given Raster widget.

  pressureDisplay = [Value2dDisplay create: self
				    setDisplayWidget: pressureRaster 
				    colormap: colormap
				    setDiscrete2dToDisplay: [modelSwarm getInfluenceGrid]];
  [pressureDisplay setDisplayMappingM: 1 C: 1]; 
 
  [pressureDisplay display];
  [pressureRaster drawSelf];

 
  opinionDisplay = [Object2dDisplay create: self
				    setDisplayWidget: opinionRaster
				    setDiscrete2dToDisplay: [modelSwarm getPositionGrid]
				    setDisplayMessage: M(drawSelfOn:)];
  //you can pass a collection to the display, but it must be a collection
  //of cells from the wrapped list grid.
  [opinionDisplay display];
  [opinionRaster drawSelf];


  // Also, tell the world raster to send mouse clicks to the heatbugDisplay
  // this allows the user to right-click on the display to probe the bugs.

  [opinionRaster setButton: ButtonRight
		 /*  Client: opinionDisplay*/ Client: [modelSwarm getPositionGrid]
		 Message: M(makeProbeAtX:Y:)];

  sprintf (hdfEZGraphName, "latane%03d.hdf", getInt (arguments,"run"));
  
  hdf5container = [HDF5 createBegin: [self getZone]];
  [hdf5container setWriteFlag: YES];
  [hdf5container  setName: hdfEZGraphName];
  hdf5container = [hdf5container createEnd];

  
  opinionGraph = [EZGraph createBegin: self];
  [opinionGraph setHDF5Container: hdf5container];
  //in hdf5 output, "File" is really the prefix used inside the file
  //name of file is value of hdf5EZGraphName
  [opinionGraph setFileName: "latane" ];
  [opinionGraph setFileOutput: 1] ;
  SET_WINDOW_GEOMETRY_RECORD_NAME (opinionGraph);
  [opinionGraph setTitle: "Opinion Trends"];
    
  [opinionGraph setAxisLabelsX: "time" Y: "Proportion"];
  opinionGraph = [opinionGraph createEnd];


  [opinionGraph setColors: ezgraphColors count: 3];
  [opinionGraph enableDestroyNotification: self
		notificationMethod: @selector (_opinionGraphDeath_:)];

  sequences[0] = [opinionGraph createSequence: "Opinion = YES"
			       withFeedFrom:  [modelSwarm getInfluenceGrid]
			       andSelector: M(getPctTypeOne)];
  
  sequences[1] = [opinionGraph createAverageSequence: "Opinion Changed"
			       withFeedFrom: [modelSwarm getCitizenList]
			       andSelector: M(getChanged)];
  

  sequences[2] =  [opinionGraph createAverageSequence: "Bolstered Minority"
				withFeedFrom: [modelSwarm getCitizenList]
				andSelector: M(getMinority)];
  
  return self;
}  

- _update_
{
 
  if (opinionRaster)
    {
      [opinionRaster erase];
      [pressureRaster erase];
      
      [pressureDisplay display];
      [pressureRaster drawSelf];
      [opinionDisplay display];
      [opinionRaster drawSelf];
     
      if (getInt (arguments, "GUIshots"))
	{
	  [self writeRasterPixmap];
	}
    }
 
  if (opinionGraph)
    {
      [opinionGraph step];
      printf("Update: percent sequence value is %f \n", [sequences[0] getCurrentValue]);
      printf("Update: change sequence value is %f \n", [sequences[1] getCurrentValue]);
    }
  return self;
}


- buildActions
{
  [super buildActions];
  
  [modelSwarm buildActions];
   
  displayActions = [ActionGroup create: self];

  [displayActions createActionTo: self message: M(_update_)];

  [displayActions createActionTo: probeDisplayManager message: M(update)];

  [displayActions createActionTo: actionCache message: M(doTkEvents)];

  displaySchedule = [Schedule createBegin: self];
  [displaySchedule setRepeatInterval: displayFrequency]; // note frequency!
  displaySchedule = [displaySchedule createEnd];
  [displaySchedule at: 0 createAction: displayActions];

  stopSchedule = [Schedule createBegin: self];
  [stopSchedule setRepeatInterval: 1]; // note frequency!
  stopSchedule = [stopSchedule createEnd];
  [stopSchedule at: 0 createActionTo: self message: M(checkToStop)];

  //Need empty displaySchedule2 to get delay checkToStop
  displaySchedule2 = [Schedule createBegin: self];
  displaySchedule2 = [displaySchedule2 createEnd];

  [displaySchedule2 at: 3 createActionTo: stopSchedule message: M(activateIn:):self];
  
  return self;
}  


- activateIn:  swarmContext
{
  [super activateIn: swarmContext];

  [modelSwarm activateIn: self];

  [displaySchedule activateIn: self];
  [displaySchedule2 activateIn: self];
  
  return [self getActivity];
}

- checkToStop
{
  double myAvg = [sequences[1] getCurrentValue];

  // printf ("average is %f \n", myAvg);
  //DBL_MIN: a secret Marcus told me about from ISO C: smallest value of double
  if ( myAvg < DBL_MIN) 
    {
      // printf ("avg equalz 0 \n");
      [self unattendedStop];
    }
  return self;
}


//cleanup if check to stop fires up (unattended gui runs)
- unattendedStop
{
  printf ("unattended stop: monitored value says it is time to quit: There were no changes in opinion\n");
  [modelSwarm finishLog];  
  
  if (getInt (arguments, "UNATTENDED"))
    {
      [modelSwarm finishLog];
      [self writeGraphPixmap];
      [self writeRasterPixmap];
      [controlPanel setStateQuit];
    }
  else
    [controlPanel setStateStopped]; 
  return self;
}

//things that need to be done to clean up after auser-click quit
- guiStop
{
  printf ("gui stop \n");
  [modelSwarm finishLog];  
  return self;
}

- writeRasterPixmap
{
  char filename[40];
  if(getCurrentTime() % displayFrequency == 0)
    {
      id pix;
      
      sprintf (filename, "opinionRaster%04dRUN%07ld.png", getInt(arguments,"run"),getCurrentTime ());
      //        sprintf (filename, "raster%07ld.png", getCurrentTime ());
      pix = [Pixmap createBegin: self];
      [pix setWidget: opinionRaster];
      [pix setDecorationsFlag: YES];
      pix = [pix createEnd]; 
      [pix save: filename] ;
      [pix drop];
    }
  return self;
}


- writeGraphPixmap
{
  id pix;
  char filename[40];
  sprintf (filename, "opinionGraph%04d.png", getInt(arguments, "run"));
  [actionCache doTkEvents];
  
  pix = [Pixmap createBegin: self];
  [pix  setWidget: [opinionGraph getGraph]];
  [pix setDecorationsFlag: NO];
  pix = [pix createEnd];
  [pix  save: filename];
  [pix drop];
  return self;
}


/*"It is necessary to drop the data writing objects in order to make
sure they finish their work.
"*/
-(void) drop
{
  printf("drop method: triggered by Swarm\n");

  [self guiStop];  //things that need to be done with user-click quit
  [opinionGraph drop];
  [super drop];
}

@end
















